/* Standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>

/* Toolkit headers */
#include <Ph.h>
#include <Pt.h>
#include <Ap.h>

// We don't need to support all the image formats Photon knows about.
#define PX_IMAGE_MODULES
#define PX_GIF_SUPPORT
#define PX_JPG_SUPPORT
#include <photon/PxImage.h>

/* Local headers */
#include "abimport.h"
#include "proto.h"
#include "playlist.h"
#include "gfuncs.h"
#include "decoder.h"

/* Number of browser pane widgets defined in the resource file. */
#define NUM_PANES 5

/* Minimum delay in ms between nav pane button presses. */
#define DEBOUNCE_MS	400

extern int autoclear;

static int current_x = 5, current_y = 5, spacing = 10;
static int highest_on_line = 0;
static int button_height = 50;
static unsigned long last_button_time = 0;

static struct history {
	char		*location;
	struct history	*next;
} *history = NULL;

enum button_type {
	BTN_UNKNOWN = 0,
	BTN_LABEL = 1,
	BTN_TRACK = 2,
	BTN_LIST = 3,
	BTN_FOLDER = 4
};

struct button_info {
	char		*label;
	char		*button_image;
	char		*font;
	char		*data;
	PgColor_t	fill_color;
	PgColor_t	text_color;
	PgColor_t	arm_color;
	int		height;
	int		width;
	int		new_line;
	enum button_type type;
	// arm image?
};

struct button {
	PtWidget_t	*widget;
	char		*data;
	struct button	*next;
};

struct file_list {
	char		**filenames;
	int		nfilenames;
};

static struct file_list cover_filenames = { NULL, 0 };

/*
 * Each browser pane instance is described by a structure.  The panes are
 * sorted from most to least recently displayed, such that panes[0] is always
 * the currently displayed pane.
 */
struct browse_pane {
	PtWidget_t	*widget;
	char		*location;
	time_t		modified;
	struct button	*buttons;
	struct button	*last_button;
};

static struct browse_pane panes[NUM_PANES];

static void update_history(void);
static void clean_pane(struct browse_pane *);
static int load_phj(char *), load_dir(char *);

/*
 * Initializes the list of panes.
 */
void init_browsepane()
{
	int		i;
	PhPoint_t	pos, *ppos;

	memset(&panes, 0, sizeof(panes));
	panes[0].widget = ABW_BrowserScrollContainer1;
	panes[1].widget = ABW_BrowserScrollContainer2;
	panes[2].widget = ABW_BrowserScrollContainer3;
	panes[3].widget = ABW_BrowserScrollContainer4;
	panes[4].widget = ABW_BrowserScrollContainer5;

	// Make sure pane 0 is the visible one
	PtWidgetToFront(panes[0].widget);

	// Move the other ones out of visible range so they don't flicker
	PtGetResource(panes[0].widget, Pt_ARG_POS, &ppos, 0);
	memcpy(&pos, ppos, sizeof(pos));

	pos.y = -20000;
	for (i = 1; i < NUM_PANES; i++)
		PtSetResource(panes[i].widget, Pt_ARG_POS, &pos, 0);
}


/*
 * Returns the current directory name.
 */
char *get_current_dir()
{
	return panes[0].location;
}


/*
 * Switches a particular pane to the foreground.
 */
static void switch_to_pane(int num)
{
	struct browse_pane	tmp;
	PhPoint_t		pos, *ppos;

	if (num == 0)
		return;

	// Scoot the other panes down by one.
	memcpy(&tmp, &panes[num], sizeof(tmp));
	memmove(&panes[1], &panes[0], sizeof(tmp) * num);
	memcpy(&panes[0], &tmp, sizeof(tmp));

	// Move the new pane 0 into place and the old one out of the way.
	PtGetResource(panes[1].widget, Pt_ARG_POS, &ppos, 0);
	memcpy(&pos, ppos, sizeof(pos));
	PtWidgetToFront(panes[0].widget);
	PtSetResource(panes[0].widget, Pt_ARG_POS, &pos, 0);
	pos.y = -20000;
	PtSetResource(panes[1].widget, Pt_ARG_POS, &pos, 0);
}


/*
 * Jumps to a new location, optionally adding the old location to history.
 */
void jump_to(char *filename, int add_to_history)
{
	struct stat		st;
	int			problem = 1;
	int			i;
	struct browse_pane	tmp;

	if (stat(filename, &st))
	{
		perror(filename);
		return;
	}

	if (add_to_history)
		update_history();
	
	set_back_button(history != NULL);

	/* If we already have a pane for this location, show it. */
	for (i = 0; i < NUM_PANES; i++)
	{
		if (panes[i].location != NULL &&
		    ! strcmp(panes[i].location, filename))
		{
			// Keep the pane list sorted properly.
			if (i > 0)
			{
				switch_to_pane(i);

				// XXX - May want to scroll to top if
				// add_to_history is true.
			}

			// We may need to reload the contents.
			if (st.st_mtime == panes[0].modified)
				return;

			clean_pane(&panes[0]);
			break;
		}
	}

	if (i >= NUM_PANES)
	{
		/* A new location.  Scoot the existing panes aside. */
		clean_pane(&panes[4]);
		switch_to_pane(4);

		panes[0].location = gstrdup(filename);
		panes[0].modified = st.st_mtime;
	}

	if (endswith(filename, ".phj"))
	{
		problem = load_phj(filename);
	}
	else
	{
		/*
		 * Unknown extension!  See if it's a directory.
		 */
		if (st.st_mode & S_IFDIR)
			problem = load_dir(filename);
		else
		{
			fprintf(stderr, "Don't know what to do with %s\n",
				filename);
		}
	}
}


/*
 * Jumps back to the previous entry in the location history.
 */
void jump_back()
{
	if (history != NULL)
	{
		struct history *oldhist;

		jump_to(history->location, 0);
		oldhist = history;
		history = oldhist->next;

		free(oldhist->location);
		free(oldhist);

		set_back_button(history != NULL);
	}
}


/*
 * Launches a playlist file.
 */
static void
launch_list(char *filename, int play)
{
	int	old_nsongs;
	int	status = 0;

	if (autoclear)
	{
		stop_playing();
		clear_list();
	}

	old_nsongs = nsongs;

	if (endswith(filename, ".m3u"))
		status = load_m3u(filename, 0);
	else if (endswith(filename, ".txt"))
		status = load_txt(filename, 0);
	if (status == 0)
	{
		order_tracks();
		if (old_nsongs == 0 && play)
			start_playing(next_track());
	}

	hide_loading_dialog();
}

/*
 * Launches a file or directory.
 */
void
launch_file(char *filename, int play)
{
	if (access(filename, R_OK))
		return;

	if (endswith(filename, ".m3u") || endswith(filename, ".txt"))
		launch_list(filename, play);
	else if (find_decoder(filename))
	{
		add_filename_to_list(filename, 1);
		set_play_controls();
		if (nsongs == 1)
		{
			order_tracks();
			start_playing(next_track());
		}
	}
	else
	{
		jump_to(gstrdup(filename), 1);
	}
}


/*
 * Handler for navigation buttons.
 */
static int
NavBtn( PtWidget_t *widget, void *apinfo, PtCallbackInfo_t *cbinfo )
{
	char	*filename;
	int	len;

	/* eliminate 'unreferenced' warnings */
	apinfo = apinfo, cbinfo = cbinfo;

	if (cbinfo != NULL && cbinfo->event != NULL)
	{
		if (cbinfo->event->timestamp < last_button_time + DEBOUNCE_MS)
			return (Pt_END);
		last_button_time = cbinfo->event->timestamp;
	}

	PtGetResource(widget, Pt_ARG_POINTER, &filename, 0);

	launch_file(filename, 1);
	return ( Pt_CONTINUE );
}


void
add_all_mp3s()
{
	struct button	*button;
	int		old_nsongs = nsongs;

	for (button = panes[0].buttons; button != NULL; button = button->next)
		if (find_decoder(button->data))
			add_filename_to_list(button->data, 0);

	set_play_controls();
	order_tracks();

	if (old_nsongs == 0)
		start_playing(next_track());
}


static void
add_button(struct button_info *info)
{
	PtCallback_t	callbacks[] = { { NavBtn, NULL } };
	PtArg_t		args[14];
	PtWidget_t	*widget;
	unsigned short	*canvas_width, *button_width, *scrollbar_width;
	unsigned short	*tmp;
	PhPoint_t	button_pos;
	struct button	*button;
	int		arg = 0;
	int		this_height;
	unsigned short	pane_width;
	PhImage_t	*image = NULL;
	int		button_type = Pt_Z_STRING;

	button = gcalloc(sizeof(struct button));
	button->data = info->data;

	PtGetResource(panes[0].widget, Pt_ARG_WIDTH,
			&canvas_width, 0);
	PtGetResource(panes[0].widget, Pt_ARG_SCROLLBAR_Y_WIDTH,
			&scrollbar_width, 0);
	pane_width = *canvas_width - *scrollbar_width;

	PtSetArg(&args[arg++], Pt_ARG_POINTER, info->data, 0);

	if (info->button_image != NULL)
	{
		image = PxLoadImage(info->button_image, NULL);

		if (image == NULL)
			perror(info->button_image);
		else
		{
			if (info->font && ! strcasecmp(info->font, "none"))
				button_type = Pt_IMAGE;
			else
				button_type = Pt_TEXT_IMAGE;
			image->flags = Ph_RELEASE_IMAGE | Ph_RELEASE_PALETTE |
					Ph_RELEASE_TRANSPARENCY_MASK |
					Ph_RELEASE_GHOST_BITMAP;
			PtSetArg(&args[arg++], Pt_ARG_LABEL_IMAGE, image, 0);
			PtSetArg(&args[arg++], Pt_ARG_LABEL_TYPE,
					button_type, 0);
			PtSetArg(&args[arg++], Pt_ARG_BALLOON_POSITION,
					Pt_BALLOON_BOTTOM, 0);
		}
	}
	else
	{
		// A regular button.  If it's one of the known types, put
		// an icon on it.  The icons are stored on some hidden
		// label widgets.

		image = NULL;

		if (info->type == BTN_UNKNOWN && info->data != NULL)
		{
			// Deduce button type based on target.
			if (find_decoder(info->data))
				info->type = BTN_TRACK;
			else if (endswith(info->data, ".m3u") ||
				 check_txt(info->data))
				info->type = BTN_LIST;
			else
				info->type = BTN_FOLDER;
		}

		switch (info->type) {
		case BTN_LIST:
			button_type = Pt_TEXT_IMAGE;
			PtGetResource(ABW_ListImage, Pt_ARG_LABEL_IMAGE,
					&image, 0);
			break;

		case BTN_FOLDER:
			button_type = Pt_TEXT_IMAGE;
			PtGetResource(ABW_FolderImage, Pt_ARG_LABEL_IMAGE,
					&image, 0);
			break;

		case BTN_TRACK:
			button_type = Pt_TEXT_IMAGE;
			PtGetResource(ABW_TrackImage, Pt_ARG_LABEL_IMAGE,
					&image, 0);
			break;

		default:
			button_type = Pt_Z_STRING;
			break;
		}

		if (image != NULL)
		{
			PhImage_t new_image;
			memcpy(&new_image, image, sizeof(PhImage_t));
			new_image.flags = Ph_USE_TRANSPARENCY;

			PtSetArg(&args[arg++], Pt_ARG_LABEL_IMAGE, &new_image,
					0);
			PtSetArg(&args[arg++], Pt_ARG_LABEL_TYPE, button_type,
					0);
			PtSetArg(&args[arg++], Pt_ARG_BALLOON_POSITION,
					Pt_BALLOON_RIGHT, 0);
			PtSetArg(&args[arg++], Pt_ARG_TEXT_IMAGE_SPACING, 5, 0);

			image = NULL;	// so we don't free it later
		}
	}
	if (button_type == Pt_Z_STRING || button_type == Pt_TEXT_IMAGE)
	{
		PtSetArg(&args[arg++], Pt_ARG_TEXT_STRING, info->label, 0);
		if (info->font != NULL)
			PtSetArg(&args[arg++], Pt_ARG_TEXT_FONT, info->font, 0);
	}
	if (info->height != 0)
		this_height = info->height;
	else
		this_height = button_height;
	PtSetArg(&args[arg++], Pt_ARG_HEIGHT, this_height, 0);
	if (info->width > 0)
		PtSetArg(&args[arg++], Pt_ARG_WIDTH, info->width, 0);
	if (info->width < 0)
		PtSetArg(&args[arg++], Pt_ARG_WIDTH, pane_width - spacing * 2, 0);
	if (info->fill_color != 0xffffffff)
		PtSetArg(&args[arg++], Pt_ARG_FILL_COLOR, info->fill_color, 0);
	else
	{
		PgColor_t	*bg;
		PtGetResource(panes[0].widget, Pt_ARG_FILL_COLOR,
				&bg, 0);
		PtSetArg(&args[arg++], Pt_ARG_FILL_COLOR, *bg, 0);
	}
	if (info->text_color != 0xffffffff)
		PtSetArg(&args[arg++], Pt_ARG_COLOR, info->text_color, 0);
	else
		PtSetArg(&args[arg++], Pt_ARG_COLOR, Pg_WHITE, 0);
	if (info->arm_color != 0xffffffff)
	{
		PtSetArg(&args[arg++], Pt_ARG_ARM_COLOR, info->text_color, 0);
		PtSetArg(&args[arg++], Pt_ARG_ARM_FILL, 1, 0);
	}
	if (info->data != NULL)
	{
		button->widget = PtCreateWidget(PtButton,
					panes[0].widget,
					arg, args);
	}
	else
	{
		button->widget = PtCreateWidget(PtLabel,
					panes[0].widget,
					arg, args);
	}
	PtRealizeWidget(button->widget);

	if (image != NULL)
		free(image);

	/* See if the button will fit on this line. */
	PtGetResource(button->widget, Pt_ARG_WIDTH, &button_width, 0);
	PtGetResource(button->widget, Pt_ARG_HEIGHT, &tmp, 0);
	this_height = *tmp;

	if (info->new_line ||
	    *button_width + current_x + spacing / 2 > pane_width)
	{
		current_x = spacing / 2;
		current_y += highest_on_line + spacing;
		PtSetResource(panes[0].widget,
				Pt_ARG_SCROLLAREA_MAX_Y,
				current_y + highest_on_line, 0);
		highest_on_line = 0;
	}

	button_pos.x = current_x;
	button_pos.y = current_y;
	PtSetResource(button->widget, Pt_ARG_POS, &button_pos, 0);

	if (*button_width > *canvas_width + spacing)
		PtSetResource(button->widget, Pt_ARG_WIDTH,
				*canvas_width - spacing, 0);

	current_x += *button_width + spacing;
	if (this_height > highest_on_line)
		highest_on_line = this_height;

	PtAddCallbacks(button->widget, Pt_CB_ACTIVATE, callbacks, 1);

	button->next = NULL;
	if (panes[0].last_button != NULL)
		panes[0].last_button->next = button;
	else
		panes[0].buttons = button;
	panes[0].last_button = button;
}


/*
 * Sets default values in a button_info structure.
 */
static void clear_info(struct button_info *info)
{
	memset(info, 0, sizeof(struct button_info));
	info->arm_color = 0xffffffff;
	info->fill_color = 0xffffffff;
	info->text_color = 0xffffffff;
	info->type = BTN_UNKNOWN;
}


/*
 * Destroys all the buttons on a pane.
 */
static void
clean_pane(struct browse_pane *pane)
{
	struct button *button, *next;

	current_x = spacing / 2;
	current_y = spacing / 2;
	highest_on_line = 0;

	PtContainerHold(pane->widget);

	for (button = pane->buttons; button != NULL; button = next)
	{
		next = button->next;
		PtDestroyWidget(button->widget);
		if (button->data != NULL)
			free(button->data);
		free(button);
	}
	pane->buttons = NULL;
	pane->last_button = NULL;

	PtSetResource(pane->widget,
			Pt_ARG_SCROLLAREA_MAX_Y, current_y + button_height, 0);
	PtContainerRelease(pane->widget);

	free(pane->location);
	pane->location = NULL;
}


/*
 * Adds the current location to the history.
 */
static void
update_history(void)
{
	struct history *hist;

	if (panes[0].location != NULL)
	{
		hist = gmalloc(sizeof(struct history));
		hist->location = gstrdup(panes[0].location);
		hist->next = history;
		history = hist;
	}
}


/*
 * Parses a color name.
 */
static PgColor_t parse_color(char *name)
{
	int red = 0, green = 0, blue = 0;

	if (isdigit(name[0]))
	{
		sscanf(name, "%d,%d,%d", &red, &green, &blue);
		return PgRGB(red, green, blue);
	}
	else
	{
		strtok(name, "\r\n");
		if (! strcasecmp(name, "white"))
			return Pg_WHITE;
		if (! strcasecmp(name, "grey"))
			return Pg_GRAY;
		if (! strcasecmp(name, "gray"))
			return Pg_GRAY;
		if (! strcasecmp(name, "black"))
			return Pg_BLACK;
		if (! strcasecmp(name, "red"))
			return Pg_RED;
		if (! strcasecmp(name, "green"))
			return Pg_GREEN;
		if (! strcasecmp(name, "blue"))
			return Pg_BLUE;
		if (! strcasecmp(name, "yellow"))
			return Pg_YELLOW;
		if (! strcasecmp(name, "magenta"))
			return Pg_MAGENTA;
		if (! strcasecmp(name, "cyan"))
			return Pg_CYAN;
		if (! strcasecmp(name, "brown"))
			return Pg_BROWN;
		if (! strcasecmp(name, "purple"))
			return Pg_PURPLE;
		if (! strcasecmp(name, "celidon"))
			return Pg_CELIDON;
	}

	printf("Unknown color '%s'", name);
	return Pg_GRAY;
}


/*
 * Parses a playlist list and creates a bunch of buttons.
 */
static int load_phj(char *filename)
{
	FILE		*fp;
	char		line[500], label[500], target[500], font[500];
	char		directory[1000], image[1000], *c;
	struct button_info info;
	int		count = 0;
	int		got_m3u = 0;
	int		got_mp3 = 0;

	c = strrchr(filename, '/');
	if (c == NULL)
		strcpy(directory, ".");
	else
	{
		*c = '\0';
		strcpy(directory, filename);
		*c = '/';
	}

	fp = fopen(filename, "r");
	if (fp == NULL)
	{
		perror(filename);
		return -1;
	}

	label[0] = '\0';
	PtBkgdHandlerProcess();
	PtContainerHold(panes[0].widget);

	set_back_button(history != NULL);
	clear_info(&info);

	while (fgets(line, sizeof(line), fp) != NULL)
	{
		// Don't make playback break up while we're building the menu
		if ((++count % 50) == 0)
		{
			PtBkgdHandlerProcess();
			if (is_playing())
				usleep(1000);
		}


		if (line[0] == ';')
			continue;

		if (line[0] == '[')
		{
			if (label[0])
				add_button(&info);

			strtok(line, "]\n\r");
			strcpy(label, line + 1);

			clear_info(&info);
			info.label = label;
			info.fill_color = Pg_DGRAY;

			continue;
		}

		if (! label[0])
			continue;
		if (! strncasecmp(line, "target=", 7))
		{
			strtok(line, "\n\r");
			if (line[7] == '/')
				strcpy(target, line + 7);
			else
				sprintf(target, "%s/%s", directory, line + 7);
			info.data = gstrdup(target);
			if (! got_m3u && (endswith(target, ".m3u") ||
					  check_txt(target)))
				got_m3u = 1;
			if (! got_mp3 && find_decoder(target))
				got_mp3 = 1;
			continue;
		}

		if (! strncasecmp(line, "font=", 5))
		{
			strtok(line, "\n\r");
			strcpy(font, line + 5);
			info.font = font;
			continue;
		}

		if (! strncasecmp(line, "height=", 7))
		{
			info.height = atoi(line + 7);
			continue;
		}

		if (! strncasecmp(line, "width=", 6))
		{
			info.width = atoi(line + 6);
			continue;
		}

		if (! strncasecmp(line, "bg=", 3))
		{
			info.fill_color = parse_color(line + 3);
			continue;
		}

		if (! strncasecmp(line, "fg=", 3))
		{
			info.text_color = parse_color(line + 3);
			continue;
		}

		if (! strncasecmp(line, "arm=", 4))
		{
			info.arm_color = parse_color(line + 4);
			continue;
		}

		if (! strncasecmp(line, "new-line=", 9))
		{
			info.new_line = atoi(line + 9);
			continue;
		}

		if (! strncasecmp(line, "image=", 6))
		{
			strtok(line, "\r\n");
			info.button_image = image;
			if (line[6] == '/')
				strcpy(image, line + 6);
			else
				sprintf(image, "%s/%s", directory, line + 6);
		}

		if (! strncasecmp(line, "type=", 5))
		{
			sscanf(line + 5, "%d", (int *)&info.type);
			continue;
		}
	}

	if (label[0])
		add_button(&info);

	PtContainerRelease(panes[0].widget);

	fclose(fp);

	set_add_tracks_menu(got_mp3);

	return 0;
}


/*
 * Adds a file to a file list.
 */
static void
add_file(char *filename, struct file_list *list)
{
	list->filenames = grealloc(list->filenames,
				sizeof(char *) * (list->nfilenames + 1));
	list->filenames[list->nfilenames++] = gstrdup(filename);
}

/*
 * Comparison function for file list sorting.
 */
static int
compare(const void *p1, const void *p2)
{
	const char **cp1 = (const char **) p1;
	const char **cp2 = (const char **) p2;

	return strcasecmp(*cp1, *cp2);
}


/*
 * Adds a cover filename to the list.
 */
void
add_cover_filename(char *filename)
{
	char	*end;

	/* Strip off whitespace and ignore empty names */
	if (filename == NULL)
		return;
	while (isspace(*filename) && *filename != '\0')
		filename++;
	if (*filename == '\0')
		return;
	end = filename + strlen(filename) - 1;
	while (isspace(*end) && end > filename)
		*end-- = '\0';
	if (end <= filename)
		return;

	add_file(filename, &cover_filenames);
}


/*
 * Converts a list of cover filenames to a comma-delimited string.
 */
char *
cover_filenames_to_text()
{
	char	*buf;
	int	i;
	int	len = 1;

	for (i = 0; i < cover_filenames.nfilenames; i++)
		len += strlen(cover_filenames.filenames[i]) + 1;

	buf = gmalloc(len);
	buf[0] = '\0';

	for (i = 0; i < cover_filenames.nfilenames; i++)
	{
		strcat(buf, cover_filenames.filenames[i]);
		if (i != cover_filenames.nfilenames - 1)
			strcat(buf, ",");
	}

	return buf;
}

/*
 * Parses a comma-delimited list of cover filenames.
 */
void
text_to_cover_filenames(char *text)
{
	char	*next, *copy;

	gfree(cover_filenames.filenames);
	cover_filenames.filenames = NULL;
	cover_filenames.nfilenames = 0;

	if (text == NULL || text[0] == '\0')
		return;

	text = copy = gstrdup(text);

	do {
		next = strchr(text, ',');
		if (next != NULL)
			*next = '\0';
		add_cover_filename(text);
		if (next == NULL)
			text = NULL;
		else
			text = next + 1;
	} while (text != NULL);
}


/*
 * Loads a directory, adding buttons for MP3s, M3Us, PHJs, and subdirectories.
 */
static int
load_dir(char *dirname)
{
	DIR		*dp;
	struct dirent	*cur;
	char		path[1000], label[500];
	int		offset;
	struct stat	st;
	struct button_info info;
	int		count = 0;
	int		i, j;
	struct file_list phj_list = {NULL, 0},
			dir_list = {NULL, 0},
			m3u_list = {NULL, 0},
			mp3_list = {NULL, 0};

	sprintf(path, "%s/", dirname);
	offset = strlen(path);

	// If the directory has an index file, use it.
	strcpy(path + offset, "index.phj");
	if (access(path, R_OK) == 0)
		return load_phj(path);

	dp = opendir(dirname);
	if (dp == NULL)
	{
		perror(dirname);
		return 1;
	}

	/*
	 * Put up a "Working" message.  This will be displayed if we process
	 * a lot of records (in which case we'll call PtBkgdHandlerProcess()
	 * and the display will be updated.)
	 */
	clear_info(&info);

	if (is_playing())
		show_loading_dialog("Loading", 0);

	PtBkgdHandlerProcess();

	set_back_button(history != NULL);

	while ((cur = readdir(dp)) != NULL)
	{
		// Update loading status
		if ((++count % 50) == 0)
			show_loading_dialog("Loading", count);

		// Skip hidden files
		if (cur->d_name[0] == '.')
			continue;

		strcpy(path + offset, cur->d_name);
		if (stat(path, &st))
		{
			perror(path);
			continue;
		}

		if (st.st_mode & S_IFDIR)
			add_file(cur->d_name, &dir_list);
		else if (endswith(cur->d_name, ".phj"))
			add_file(cur->d_name, &phj_list);
		else if (endswith(cur->d_name, ".m3u") ||
			 check_txt(path))
			add_file(cur->d_name, &m3u_list);
		else if (find_decoder(cur->d_name))
			add_file(cur->d_name, &mp3_list);
	}

	closedir(dp);

	qsort(dir_list.filenames, dir_list.nfilenames, sizeof(char *),
		compare);
	qsort(phj_list.filenames, phj_list.nfilenames, sizeof(char *),
		compare);
	qsort(m3u_list.filenames, m3u_list.nfilenames, sizeof(char *),
		compare);
	qsort(mp3_list.filenames, mp3_list.nfilenames, sizeof(char *),
		compare);

	/*
	 * Don't update incrementally while we build the list; it'll just
	 * slow things down.
	 */
	PtContainerHold(panes[0].widget);

	count = 0;

	for (i = 0; i < phj_list.nfilenames; i++)
	{
		if ((++count % 10) == 0)
			show_loading_dialog("Processing...", count);

		clear_info(&info);
		strcpy(label, phj_list.filenames[i]);
		label[strlen(label) - 4] = '\0';
		fs_to_utf8(label, sizeof(label));
		strcpy(path + offset, phj_list.filenames[i]);
		info.data = gstrdup(path);
		info.label = label;
		info.font = "lub12b";
		info.fill_color = Pg_DBLUE;
		info.type = BTN_FOLDER;
		add_button(&info);
		free(phj_list.filenames[i]);
	}

	for (i = 0; i < dir_list.nfilenames; i++)
	{
		int	subdir_offset;

		if ((++count % 10) == 0)
			show_loading_dialog("Processing...", count);

		clear_info(&info);
		strcpy(label, dir_list.filenames[i]);
		fs_to_utf8(label, sizeof(label));
		strcpy(path + offset, dir_list.filenames[i]);
		info.data = gstrdup(path);
		info.label = label;
		info.font = "lub12b";
		info.fill_color = Pg_DBLUE;
		info.text_color = Pg_YELLOW;
		info.type = BTN_FOLDER;

		if (cover_filenames.nfilenames > 0)
		{
			// If there's an album image, use it.
			subdir_offset = strlen(path);
			path[subdir_offset++] = '/';

			for (j = 0; j < cover_filenames.nfilenames; j++)
			{
				strcpy(path + subdir_offset,
					cover_filenames.filenames[j]);
				if (access(path, R_OK) == 0)
				{
					info.button_image = path;
					break;
				}
			}
		}

		add_button(&info);
		free(dir_list.filenames[i]);
	}

	for (i = 0; i < m3u_list.nfilenames; i++)
	{
		if ((++count % 10) == 0)
			show_loading_dialog("Processing...", count);

		clear_info(&info);
		strcpy(label, m3u_list.filenames[i]);
		label[strlen(label) - 4] = '\0';
		fs_to_utf8(label, sizeof(label));
		strcpy(path + offset, m3u_list.filenames[i]);
		info.data = gstrdup(path);
		info.label = label;
		info.type = BTN_LIST;
		info.fill_color = Pg_DGRAY;
		info.text_color = Pg_WHITE;
		add_button(&info);
		free(m3u_list.filenames[i]);
	}

	for (i = 0; i < mp3_list.nfilenames; i++)
	{
		if ((++count % 10) == 0)
			show_loading_dialog("Processing...", count);

		clear_info(&info);
		strcpy(label, mp3_list.filenames[i]);
		label[strlen(label) - 4] = '\0';
		fs_to_utf8(label, sizeof(label));
		strcpy(path + offset, mp3_list.filenames[i]);
		info.data = gstrdup(path);
		info.label = label;
		info.font = "TextFont07";
		info.fill_color = Pg_DGREEN;
		info.height = 30;
		info.width = -1;
		info.type = BTN_TRACK;
		add_button(&info);
		free(mp3_list.filenames[i]);
	}

	set_add_tracks_menu(mp3_list.filenames != NULL);

	if (phj_list.filenames != NULL)
		free(phj_list.filenames);
	if (dir_list.filenames != NULL)
		free(dir_list.filenames);
	if (m3u_list.filenames != NULL)
		free(m3u_list.filenames);
	if (mp3_list.filenames != NULL)
		free(mp3_list.filenames);

	PtContainerRelease(panes[0].widget);
	hide_loading_dialog();

	return 0;
}
