/*
 * Functions for translating to and from Unicode UTF-8 encoding.  Photon
 * uses UTF-8 internally and any non-ASCII characters we want to display
 * have to be represented that way.
 *
 * We support two different character sets for external data.  The
 * native character set is used to encode file contents, and the filesystem
 * character set is used to encode file and directory names.  This is
 * necessary for Windows systems, which (in the US) use ISO-8859-1 (Latin-1)
 * for file contents, but use the Windows native codepage-1252 character
 * set in their filenames.
 */
#include <stdio.h>
#include <photon/PxProto.h>
#include "gfuncs.h"

/* The file server's character set for data. */
char *native_charset = "ISO-8859-1";

/* The file server's character set for filenames. */
char *fs_charset = "cp437";

/* Translation control structures for Photon translation libraries. */
static struct PxTransCtrl *native_ctrl = NULL;
static struct PxTransCtrl *fs_ctrl = NULL;


/*
 * Translates a string in-place from native to UTF-8.
 */
int
native_to_utf8(char *str, int maxlen)
{
	int	len;
	int	dummy;
	char	*buf = gmalloc(maxlen);

	if (native_ctrl == NULL)
	{
		native_ctrl = PxTranslateSet(NULL, native_charset);
		if (native_ctrl == NULL)
		{
			perror(native_charset);
			return -1;
		}
	}

	if (PxTranslateToUTF(native_ctrl, str, strlen(str), &dummy,
				buf, maxlen - 1, &len))
	{
		perror("translating native to UTF");
		return -1;
	}

	buf[len] = '\0';
	strcpy(str, buf);
	free(buf);
	return 0;
}


/*
 * Translates a string in-place from filesystem charset to UTF-8.
 */
int
fs_to_utf8(char *str, int maxlen)
{
	int	len;
	int	dummy;
	char	*buf = gmalloc(maxlen);

	if (fs_ctrl == NULL)
	{
		fs_ctrl = PxTranslateSet(NULL, fs_charset);
		if (fs_ctrl == NULL)
		{
			perror(fs_charset);
			return -1;
		}
	}

	if (PxTranslateToUTF(fs_ctrl, str, strlen(str), &dummy,
				buf, maxlen - 1, &len))
	{
		perror("translating fs to UTF");
		return -1;
	}

	buf[len] = '\0';
	strcpy(str, buf);
	free(buf);
	return 0;
}


/*
 * Translates a string from UTF8 to the filesystem charset.
 */
int
utf8_to_native(char *str, int maxlen)
{
	char	*buf = gmalloc(maxlen);
	int	dummy, len;

	if (native_ctrl == NULL)
	{
		native_ctrl = PxTranslateSet(NULL, native_charset);
		if (native_ctrl == NULL)
			return -1;
	}

	if (PxTranslateFromUTF(native_ctrl, str, strlen(str), &dummy,
				buf, maxlen - 1, &len))
	{
		perror("translating UTF to native");
		return -1;
	}

	buf[len] = '\0';
	strcpy(str, buf);
	free(buf);
	return 0;
}


/*
 * Translates a string from UTF8 to the filesystem charset.
 */
int
utf8_to_fs(char *str, int maxlen)
{
	char	*buf = gmalloc(maxlen);
	int	dummy, len;

	if (fs_ctrl == NULL)
	{
		fs_ctrl = PxTranslateSet(NULL, fs_charset);
		if (fs_ctrl == NULL)
			return -1;
	}

	if (PxTranslateFromUTF(fs_ctrl, str, strlen(str), &dummy,
				buf, maxlen - 1, &len))
	{
		perror("translating UTF to fs");
		return -1;
	}

	buf[len] = '\0';
	strcpy(str, buf);
	free(buf);
	return 0;
}
