/*
 * Module to talk to madplay.
 */
#include <stdio.h>
#include <unix.h>
#include <ctype.h>
#include <sys/resource.h>
#include <unistd.h>
#include <ioctl.h>
#include <signal.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include "ablibs.h"
#include "abimport.h"
#include "abdefine.h"
#include "playlist.h"
#include "decoder.h"

static int madplay_fd = -1;
static int madplay_pid = -1;
static int paused = 0;
static FILE *madplay_file = NULL;

extern int use_replaygain;
extern double add_decibels;

static void set_text(PtWidget_t *widget, char const *str)
{
	PtSetResource(widget, Pt_ARG_TEXT_STRING, str, strlen(str));
}


/*
 * Stops playing an MP3 file.  Waits for madplay to exit if it's still
 * running.
 */
static void mp_stop_playing()
{
	if (madplay_fd < 0)
		return;

	PtAppRemoveFd(NULL, madplay_fd);
	close(madplay_fd);	// this will SIGHUP the child
	if (paused)
		kill(madplay_pid, SIGCONT);
	kill(madplay_pid, SIGINT);
	wait(NULL);
	madplay_fd = -1;
	madplay_pid = -1;
	paused = 0;
}


/*
 * Parses a line from madplay and updates the UI accordingly.
 *
 * Input:
 *	Line to parse
 */
static void
parse_line(char *line)
{
	if (line[0] == '>')
	{
		/* Should never get this! */
		stop_playing();
		return;
	}

	if (line[3] == ':')
	{
		void		(*func)(int);
		int		seconds;

		// Our modified madplay spits out current and total time.

		seconds = atoi(line + 7) + atoi(line + 4) * 60;
		set_current_time(seconds);

		seconds = atoi(line + 17) + atoi(line + 14) * 60;
		set_total_time(seconds);
		return;
	}

	if (isdigit(line[0]) && strstr(line, "frames decoded"))
	{
		track_is_done();
		return;
	}

	while (line[0] != '\0' && isspace(line[0]))
		line++;
	if (strchr(line, ':'))
	{
		char	*keyword = strtok(line, ":");
		char	*value = strtok(NULL, "\n\r");

		while (value[0] != '\0' && isspace(value[0]))
			value++;

		if (! strcmp(keyword, "Title"))
			set_track_name(value);
		else if (! strcmp(keyword, "Artist"))
			set_artist_name(value);
		else if (! strcmp(keyword, "Album"))
			set_album_name(value);
		return;
	}
}


/*
 * Writes a command to madplay.
 */
static void
write_pipe(char cmd)
{
	char c = cmd;

	write(madplay_fd, &c, 1);
}


/*
 * Toggles pause mode.
 */
static void mp_toggle_pause()
{
	if (madplay_fd >= 0)
	{
		if (paused)
			kill(madplay_pid, SIGCONT);
		else
			kill(madplay_pid, SIGSTOP);

		paused = ! paused;
	}
}


/*
 * Reads a line from madplay and updates the UI accordingly.  This is
 * a callback from Photon, called when there's input on the madplay
 * file descriptor.
 *
 * Input:
 *	File descriptor (unused here)
 *	App-supplied data (unused here)
 *	Read/write flag (unused here)
 */
static int
madplay_read(int fd, void *data, unsigned mode)
{
	char	buf[500];
	int	i = 0;
	int	c;

	/*
	 * Read the line a character at a time since madplay terminates
	 * lines with \r rather than \n.
	 */
	do {
		c = fgetc(madplay_file);
		if (c == '\r' || c == '\n' || c == EOF)
			buf[i] = '\0';
		else
			buf[i] = c;
	} while (buf[i++] && i < sizeof(buf) - 1);
	if (i == sizeof(buf) - 1)
		buf[i] = '\0';

	if (buf[0] != '\0')
		parse_line(buf);

	// If madplay died, move on.
	if (c == EOF)
		track_is_done();

	return 0;
}


/*
 * Returns true if madplay is playing.
 */
static int mp_is_playing()
{
	return madplay_fd >= 0;
}


/*
 * Starts playing a playlist.
 *
 * Input:
 *	Name of file to start playing.
 */
static int mp_start_playing(char *filename)
{
	char	*args[8];
	int	i;
	int	argc = 0;
	char	ttyname[50];
	char	db[20];
	char	scratch[100];
	int	pty_slave_fd;
	int	tty_fd;
	int	fds[3];
	int	socks[2];
	struct inheritance inh;

	if (madplay_fd >= 0)
		mp_stop_playing();

	if (filename == NULL)
		return -1;

	pipe(socks);

	args[argc++] = "madplay";
	args[argc++] = "-v";
	if (use_replaygain)
	{
		args[argc++] = "-Ga";
		sprintf(db, "%.2f", add_decibels);
		args[argc++] = db;
	}
	args[argc++] = filename;
	args[argc++] = NULL;
	fds[0] = socks[1];
	fds[1] = socks[1];
	fds[2] = socks[1];
	memset(&inh, 0, sizeof(inh));
	inh.flags = SPAWN_CHECK_SCRIPT | SPAWN_SEARCH_PATH | SPAWN_SETGROUP;
	inh.pgroup = SPAWN_NEWPGROUP;

	madplay_pid = spawn(args[0], 3, fds, &inh, args, environ);
	if (madplay_pid < 0)
	{
		spawn_error(args[0]);
		return -1;
	}

	close(socks[1]);
	madplay_fd = socks[0];
	madplay_file = fdopen(madplay_fd, "r");

	// Read the madplay banner line to make sure it started up okay.
	if (fgets(scratch, sizeof(scratch), madplay_file) == NULL)
	{
		perror("reading from madplay");
		stop_playing();
		madplay_fd = -1;
		return -1;
	}
	PtAppAddFd(NULL, madplay_fd, Pt_FD_READ, madplay_read, NULL);

	paused = 0;

	return 0;
}


static char *file_types[] = { ".mp3", ".mp2", NULL };
struct decoder decoder_madplay = {
	"madplay",
	file_types,
	mp_stop_playing,
	mp_start_playing,
	mp_is_playing,
	mp_toggle_pause,
	NULL
};
