/*
 * Module to talk to ogg123 using a pseudo-tty.
 */
#include <stdio.h>
#include <unix.h>
#include <ctype.h>
#include <sys/resource.h>
#include "ablibs.h"
#include "abimport.h"
#include "abdefine.h"
#include "playlist.h"
#include "decoder.h"

static int player_fd = -1;
static int player_pid = -1;
static FILE *player_file = NULL;
static int paused = 0;


static void set_text(PtWidget_t *widget, char const *str)
{
	PtSetResource(widget, Pt_ARG_TEXT_STRING, str, strlen(str));
}


/*
 * Stops playing an MP3 file.  Waits for ogg123 to exit if it's still
 * running.
 */
static void og_stop_playing()
{
	if (player_fd < 0)
		return;

	PtAppRemoveFd(NULL, player_fd);
	close(player_fd);	// this will SIGHUP the child
	if (paused)
		kill(player_pid, SIGCONT);
	kill(player_pid, SIGINT);
	wait(NULL);
	player_fd = -1;
	player_pid = -1;
	paused = 0;
}


/*
 * Parses a line from ogg123 and updates the UI accordingly.
 *
 * Input:
 *	Line to parse
 */
static void
parse_line(char *line)
{
	strtok(line, "\r\n");

	if (! strncmp(line, "Time: ", 6))
	{
		int	seconds;

		seconds = atoi(line + 6) * 60 + atoi(line + 9);
		set_current_time(seconds);
		return;
	}

	if (! strcmp(line, "Done."))
	{
		track_is_done();
		return;
	}

	while (line[0] != '\0' && isspace(line[0]))
		line++;
	if (strchr(line, ':'))
	{
		char	*keyword = strtok(line, ":");
		char	*value = strtok(NULL, "\n\r");

		while (value[0] != '\0' && isspace(value[0]))
			value++;

		if (! strcmp(keyword, "Title"))
			set_track_name(value);
		else if (! strcmp(keyword, "Artist"))
			set_artist_name(value);
		else if (! strcmp(keyword, "Album"))
			set_album_name(value);
		else if (! strcmp(keyword, "Total Seconds"))
			set_total_time(atoi(value));
		return;
	}
}


/*
 * Toggles pause mode.
 */
static void og_toggle_pause()
{
	paused = ! paused;

	if (paused)
		kill(player_pid, SIGSTOP);
	else
		kill(player_pid, SIGCONT);
}


/*
 * Reads a line from madplay and updates the UI accordingly.  This is
 * a callback from Photon, called when there's input on the madplay
 * file descriptor.
 *
 * Input:
 *	File descriptor (unused here)
 *	App-supplied data (unused here)
 *	Read/write flag (unused here)
 */
static int
player_read(int fd, void *data, unsigned mode)
{
	char	buf[500];
	int	i = 0;
	int	c;

	/*
	 * Read the line a character at a time since player terminates
	 * lines with \r rather than \n.
	 */
	do {
		c = fgetc(player_file);
		if (c == '\r' || c == '\n' || c == EOF)
			buf[i] = '\0';
		else
			buf[i] = c;
	} while (buf[i++] && i < sizeof(buf) - 1);
	if (i == sizeof(buf) - 1)
		buf[i] = '\0';

	if (buf[0] != '\0')
		parse_line(buf);

	// If player died, move on.
	if (c == EOF)
		track_is_done();

	return 0;
}


/*
 * Returns true if player is playing.
 */
static int og_is_playing()
{
	return player_fd >= 0;
}


/*
 * Starts playing a playlist.
 *
 * Input:
 *	Name of file to start playing.
 */
static int og_start_playing(char *filename)
{
	char	*args[4];
	int	pipefds[2];
	int	fds[3];
	int	i;
	char	ttyname[50];
	char	scratch[100];
	struct inheritance inh;

	if (player_fd >= 0)
		og_stop_playing();

	if (filename == NULL)
		return -1;

	pipe(pipefds);

	args[0] = "ogg123";
	args[1] = "-v";
	args[2] = filename;
	args[3] = NULL;
	fds[0] = pipefds[1];
	fds[1] = pipefds[1];
	fds[2] = pipefds[1];
	memset(&inh, 0, sizeof(inh));
	inh.flags = SPAWN_CHECK_SCRIPT | SPAWN_SEARCH_PATH;

	player_pid = spawn(args[0], 3, fds, &inh, args, environ);
	if (player_pid < 0)
	{
		spawn_error(args[0]);
		return -1;
	}

	close(pipefds[1]);
	player_fd = pipefds[0];
	player_file = fdopen(player_fd, "r");

	// Read the ogg123 banner line to make sure it started up okay.
	if (fgets(scratch, sizeof(scratch), player_file) == NULL)
	{
		perror("reading from player");
		stop_playing();
		player_fd = -1;
		return -1;
	}

	paused = 0;

	parse_line(scratch);
	PtAppAddFd(NULL, player_fd, Pt_FD_READ, player_read, NULL);

	return 0;
}


static char *file_types[] = { ".ogg", NULL };
struct decoder decoder_ogg123 = {
	"ogg123",
	file_types,
	og_stop_playing,
	og_start_playing,
	og_is_playing,
	og_toggle_pause,
	NULL
};
