/*
 * Guarded memory allocation, string handling functions.
 */
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "gfuncs.h"

#define DEFAULT_SIZE	65536

struct mem_arena {
	char	*buf;		// Alloced buffer
	int	size;		// Buffer capacity
	int	pos;		// Next byte that can be returned to user
	struct mem_arena *next;
};


static void *
guard(void *result)
{
	if (result == NULL)
	{
		fprintf(stderr, "Out of memory!  Aborting.\n");
		abort();
	}

	return result;
}


void *
gmalloc(int size)
{
	return guard(malloc(size));
}

void *
grealloc(void *orig, int size)
{
	if (orig == NULL)
		return gmalloc(size);
	else
		return guard(realloc(orig, size));
}

char *
gstrdup(char *str)
{
	if (str == NULL)
		return NULL;

	return (char *) guard(strdup(str));
}

void *
gcalloc(int size)
{
	return guard(calloc(1, size));
}

int
gstrlen(char *str)
{
	if (str == NULL)
		return 0;
	return strlen(str);
}

void
gfree(void *ptr)
{
	if (ptr != NULL)
		free(ptr);
}

int
endswith(const char *str, const char *suffix)
{
	int stringlen, suffixlen;

	if (str == NULL || suffix == NULL ||
	    (stringlen = strlen(str)) < (suffixlen = strlen(suffix)))
	{
		return 0;
	}
	return ! strcasecmp(str + stringlen - suffixlen, suffix);
}

void
gstrncpy(char *dest, char *src, int len)
{
	strncpy(dest, src, len - 1);
	dest[len - 1] = '\0';
}


/*
 * Arena-style memory allocation package.  Used when we want to allocate a
 * bunch of related memory that won't be realloced or freed except as a
 * group.  We can thus avoid the time and space overhead of heap management
 * and just hand out slices of a big preallocated buffer.
 */

void *
arena_create(int initial_size)
{
	struct mem_arena	*arena = gmalloc(sizeof(struct mem_arena));

	if (initial_size == 0)
		initial_size = DEFAULT_SIZE;
	arena->buf = gmalloc(initial_size);
	arena->size = initial_size;
	arena->pos = 0;
	arena->next = NULL;

	return arena;
}

void *
arena_malloc(void *varena, int size)
{
	struct mem_arena	*arena = (struct mem_arena *) varena;
	struct mem_arena	*new_arena;
	struct mem_arena	temp_arena;
	int			new_arena_size;
	void			*ptr;

	if (arena == NULL)
		return NULL;
	if (arena->pos + size > arena->size)
	{
		/*
		 * No more space in this arena.  Allocate a new one.  If
		 * we're being asked for a really big block, give it its
		 * own arena and continue using the current one; otherwise
		 * put the new arena at the head of the list so we'll
		 * use it next time.
		 */
		if (size > arena->size)
			new_arena_size = size;
		else
			new_arena_size = arena->size;

		new_arena = arena_create(new_arena_size);
		new_arena->next = arena->next;
		arena->next = new_arena;

		if (size > arena->size)
		{
			new_arena->pos = size;
			return new_arena->buf;
		}

		// Swap mem_arena contents so the user can keep using
		// his old pointer.
		memcpy(&temp_arena, arena, sizeof(struct mem_arena));
		arena->buf = new_arena->buf;
		arena->pos = new_arena->pos;
		arena->size = new_arena->size;
		new_arena->buf = temp_arena.buf;
		new_arena->pos = temp_arena.pos;
		new_arena->size = temp_arena.size;
	}

	ptr = &arena->buf[arena->pos];
	arena->pos += size;
	return ptr;
}

void
arena_destroy(void **varena)
{
	struct mem_arena	*arena = (struct mem_arena *) *varena;
	struct mem_arena	*next_arena;

	while (arena != NULL)
	{
		next_arena = arena->next;
		free(arena->buf);
		free(arena);
		arena = next_arena;
	}

	*varena = NULL;
}

char *
arena_strdup(void *varena, char *str)
{
	char	*ptr = arena_malloc(varena, strlen(str) + 1);

	strcpy(ptr, str);
	return ptr;
}

void *
arena_calloc(void *varena, int size)
{
	char	*ptr = arena_malloc(varena, size);

	memset(ptr, 0, size);
	return ptr;
}
