/*
 * Manipulation of M3U (WinAmp playlist) files.
 */
#include <stdio.h>
#include <string.h>
#include "playlist.h"
#include "gfuncs.h"
#include "decoder.h"

struct m3u {
	char	*filename;
	char	*directory;
	int	directory_len;
	FILE	*fp;
};


/*
 * Reads an M3U file and adds all its records to the playlist.
 *
 * Returns:
 *	0	Playlist loaded successfully
 *	1	Couldn't load list.
 */
int
load_m3u(char *filename, int update_widget)
{
	char	*c;
	char	directory[1000];
	char	line[1000], title[1000];
	FILE	*fp;
	int	count = 0;

	title[0] = '\0';

	c = strrchr(filename, '/');
	if (c == NULL)
		strcpy(directory, ".");
	else
	{
		*c = '\0';
		strcpy(directory, filename);
		*c = '/';
	}

	fp = fopen(filename, "r");
	if (fp == NULL)
	{
		perror(filename);
		return 1;
	}

	/*
	 * Each line in an M3U file is either a comment or a filename.
	 * WinAmp writes special comments with additional information
	 * about tracks, so we preserve that data to feed to the playlist
	 * code when we get to the filename.
	 */
	while (fgets(line, sizeof(line), fp) != NULL)
	{
		/* Strip trailing \r and/or \n. */
		strtok(line, "\r\n");

		if (! strncmp(line, "#EXTINF:", 8) && strchr(line, ','))
		{
			strtok(line, ",");
			c = strtok(NULL, "");
			if (c != NULL)
			{
				strcpy(title, c);
				/* Must use UTF-8 in Photon labels... */
				native_to_utf8(title, sizeof(title));
			}
		}
		else if (line[0] == '#')
			continue;
		else
		{
			if ((++count % 100) == 0)
				show_loading_dialog("Loading", count);

			if (! find_decoder(line))
				continue;

			/* Convert Windows relative paths to UNIX format */
			for (c = line; *c != '\0'; c++)
				if (*c == '\\')
					*c = '/';

			/* Convert to local mount points */
			map_dir(line);

			if (title[0] == '\0')	// no title in M3U
			{
				// Construct title out of filename
				c = strrchr(line, '/');
				if (c == NULL)
					strcpy(title, line);
				else
					strcpy(title, c + 1);

				// Strip off extension
				c = strrchr(title, '.');
				if (c != NULL)
					*c = '\0';

				// Translate accented characters
				native_to_utf8(title, sizeof(title));
			}

			add_to_list(directory, line, title, update_widget);

			/* Clear out auxiliary information for next track */
			title[0] = '\0';
		}
	}

	fclose(fp);
	set_play_controls();

	return (count == 0);
}


/*
 * Opens an M3U for writing.
 *
 * Input:
 *	Directory to write to
 *	File to open
 * Returns:
 *	Handle to M3U, or NULL on error.
 */
void *
m3u_create(char *directory, char *filename)
{
	struct m3u	*m3u = gmalloc(sizeof(struct m3u));
	char		path[1000];

	sprintf(path, "%s/%s", directory, filename);
	if (strchr(filename, '.') == NULL)
		strcat(path, ".m3u");
	
	m3u->fp = fopen(path, "w");
	if (m3u->fp == NULL)
	{
		gfree(m3u);
		return NULL;
	}

	fputs("#EXTM3U\r\n", m3u->fp);

	m3u->directory = gstrdup(directory);
	m3u->directory_len = strlen(directory);
	m3u->filename = gstrdup(filename);

	return m3u;
}


/*
 * Writes an entry to an M3U.
 *
 * Input:
 *	Handle to M3U as returned by create_m3u()
 *	Title of track to write
 *	Path to track
 * Returns:
 *	0 on success, -1 on error.
 */
int
m3u_write(void *handle, char *title, char *path)
{
	struct m3u	*m3u = (struct m3u *)handle;
	char		native_path[1000];
	char		native_title[1000];
	char		*filename;

	// If this file is under the directory we're saving the playlist to,
	// write a relative path rather than the entire path.
	if (! strncmp(path, m3u->directory, m3u->directory_len))
		path += m3u->directory_len + 1;

	strcpy(native_path, path);
	fs_to_utf8(native_path, sizeof(native_path));

	// If the title differs from the filename, write it out so we'll
	// remember it for next time.
	filename = strrchr(native_path, '/');
	if (filename == NULL)
		filename = native_path;
	else
		filename++;
	if (strncmp(title, filename, strlen(title)) ||
	    filename[strlen(title)] != '.')
	{
		strcpy(native_title, title);
		utf8_to_native(native_title, sizeof(native_title));
		if (fprintf(m3u->fp, "#EXTINF:-1,%s\r\n", native_title) < 14)
			return -1;
	}

	if (utf8_to_native(native_path, sizeof(native_path)))
	{
		if (utf8_to_fs(native_path, sizeof(native_path)))
		{
			return -1;
		}
	}

	if (fputs(native_path, m3u->fp) == EOF ||
	    fputc('\r', m3u->fp) == EOF ||
	    fputc('\n', m3u->fp) == EOF)
	{
		return -1;
	}

	return 0;
}


/*
 * Closes an M3U.
 *
 * Input:
 *	Handle to M3U as returned by create_m3u()
 */
void
m3u_close(void *handle)
{
	struct m3u *m3u = (struct m3u *)handle;

	gfree(m3u->filename);
	gfree(m3u->directory);
	fclose(m3u->fp);
	gfree(m3u);
}
