/*
 * Miscellaneous utility functions.
 */
#include <errno.h>
#include <Pt.h>
#include "ablibs.h"
#include "abimport.h"
#include "abdefine.h"
#include "playlist.h"
#include "decoder.h"
#include "gfuncs.h"

extern int slider_dragging;
extern int nsongs;

char	current_artist[1000];
char	current_album[1000];
char	current_track[1000];
int	current_time;
int	current_length;

int	loading_is_shown = 0;
PhPoint_t loading_pos;
PgColor_t enabled_button_color = Pg_BLACK;
PgColor_t disabled_button_color = Pg_MGRAY;

static int paused = 0;

#define PREV_BUTTON 0
#define PLAY_BUTTON 2
#define PAUSE_BUTTON 4
#define NEXT_BUTTON 6
#define STOP_BUTTON 8

PhImage_t *control_images[10];

/*
 * Disable a ghostable widget.
 */
void
disable_widget(PtWidget_t *widget)
{
	PtSetResource(widget, Pt_ARG_FLAGS, Pt_TRUE, Pt_BLOCKED | Pt_GHOST);
}

/*
 * Enable a ghostable widget.
 */
void
enable_widget(PtWidget_t *widget)
{
	PtSetResource(widget, Pt_ARG_FLAGS, Pt_FALSE, Pt_BLOCKED | Pt_GHOST);
}

/*
 * Enable an image button (ghosting doesn't work on Audrey).
 */
void
enable_control(PtWidget_t *widget, int index)
{
	PtSetResource(widget, Pt_ARG_LABEL_IMAGE, control_images[index + 1], 0);
	PtSetResource(widget, Pt_ARG_FLAGS, Pt_FALSE, Pt_BLOCKED);
}

/*
 * Disable an image button (ghosting doesn't work on Audrey).
 */
void
disable_control(PtWidget_t *widget, int index)
{
	PtSetResource(widget, Pt_ARG_LABEL_IMAGE, control_images[index], 0);
	PtSetResource(widget, Pt_ARG_FLAGS, Pt_TRUE, Pt_BLOCKED);
}


/*
 * Convenience function to set a widget's text.
 */
static void set_text(PtWidget_t *widget, char const *str)
{
	PtSetResource(widget, Pt_ARG_TEXT_STRING, str, 0);
}


/*
 * Given a filename, figure out which decoder to use.
 */
struct decoder *find_decoder(char *filename)
{
	int	dec, ext;

	for (dec = 0; decoders[dec] != NULL; dec++)
		for (ext = 0; decoders[dec]->extensions[ext] != NULL; ext++)
			if (endswith(filename, decoders[dec]->extensions[ext]))
				return decoders[dec];
	return NULL;
}


/*
 * Frontend functions for the decoder entry points.  The frontends take
 * care of miscellaneous UI cleanup as well.
 */
int start_playing(char *filename)
{
	char		buf[1000];
	struct decoder *dec;
	int		ret = -1;

	stop_playing();
	if (filename == NULL)
		return -1;

	if ((dec = find_decoder(filename)) == NULL)
		return -1;
	decoder = dec;

	set_text(ABW_TrackLabel, "");
	set_text(ABW_ArtistLabel, "");
	set_text(ABW_AlbumLabel, "");

	/*
	 * See if the file exists...
	 */
	if (access(filename, R_OK) == 0)
		ret = decoder->start_playing(filename);
	else
	{
		/*
		 * Nope.  Try converting character sets, since we might have
		 * gotten the filename from an M3U, which could use a different
		 * character set than the filesystem.
		 */
		perror(filename);
		strcpy(buf, filename);
		native_to_utf8(buf, sizeof(buf));
		utf8_to_fs(buf, sizeof(buf));
		if (access(buf, R_OK) == 0)
			ret = decoder->start_playing(buf);
		else
		{
			printf("File %s not found (also tried %s)\n",
				filename, buf);
		}
	}

	set_play_controls();
	paused = 0;

	return ret;
}

void stop_playing()
{
	if (decoder->is_playing())
	{
		decoder->stop_playing();
		set_text(ABW_TrackLabel, "");
		set_text(ABW_ArtistLabel, "");
		set_text(ABW_AlbumLabel, "");
	}

	set_play_controls();
	paused = 0;
}

void toggle_pause()
{
	if (decoder->is_playing())
		decoder->toggle_pause();
	paused = ! paused;
}

int is_playing()
{
	return decoder->is_playing();
}

int is_paused()
{
	return paused;
}

void jump_to_time(int seconds)
{
	if (decoder->jump_to_time != NULL)
		decoder->jump_to_time(seconds);
}

void spawn_error(const char *program)
{
	char	*path = getenv("PATH");

	perror(program);

	switch (errno) {
	case ENOENT:
		fputs("Looked in the following directories:\n\t", stderr);
		while (path != NULL && *path != '\0')
		{
			if (*path == ':')
			{
				fputc('\n', stderr);
				fputc('\t', stderr);
			}
			else
				fputc(*path, stderr);
			path++;
		}
		fprintf(stderr, "\nMake sure %s is in one of them and is "
			"executable.\n", program);
		break;

	case EACCES:
		fprintf(stderr, "Make sure %s is executable (run \"chmod +x\" "
			"on it.)\n", program);
		break;
	}
}


/*
 * Callbacks from the decoder code.
 */
void track_is_done()
{
	char	*track;

	do {
		track = next_track();
		if (start_playing(track) == 0)
			break;
	} while (track != NULL);
}

void set_track_name(char *name)
{
	char	buf[1000];

	gstrncpy(current_track, name, sizeof(current_track));
	strncpy(buf, name, sizeof(buf));
	native_to_utf8(buf, sizeof(buf));
	set_text(ABW_TrackLabel, buf);
}

void set_album_name(char *name)
{
	char	buf[1000];

	gstrncpy(current_album, name, sizeof(current_album));
	strncpy(buf, name, sizeof(buf));
	native_to_utf8(buf, sizeof(buf));
	set_text(ABW_AlbumLabel, buf);
}

void set_artist_name(char *name)
{
	char	buf[1000];

	gstrncpy(current_artist, name, sizeof(current_artist));
	strncpy(buf, name, sizeof(buf));
	native_to_utf8(buf, sizeof(buf));
	set_text(ABW_ArtistLabel, buf);
}

void set_total_time(int seconds)
{
	char	scratch[5];

	current_length = seconds;
	sprintf(scratch, "%02d:%02d", seconds / 60, seconds % 60);
	set_text(ABW_LengthLabel, scratch);
	if (! slider_dragging)
		PtSetResource(ABW_TimeSlider, Pt_ARG_MAXIMUM, (long) seconds, 0);
}

void set_current_time(int seconds)
{
	char	scratch[5];

	current_time = seconds;
	sprintf(scratch, "%02d:%02d", seconds / 60, seconds % 60);
	set_text(ABW_TimeLabel, scratch);
	PtSetResource(ABW_TimeSlider, Pt_ARG_GAUGE_VALUE, (long) seconds, 0);
}


/*
 * Enables the "Back" button if there's someplace for it to point.
 */
void
set_back_button(int enabled)
{
	if (enabled)
		enable_widget(ABW_BackButton);
	else
		disable_widget(ABW_BackButton);
}


/*
 * Enables the "Add" button if there are MP3s to add.
 */
void
enable_menu_item(int item, int enabled)
{
	ApModifyItemState(&CommandMenu,
			enabled ? AB_ITEM_NORMAL : AB_ITEM_DIM,
			item, NULL);
}


/*
 * Enables the "Add" button if there are MP3s to add.
 */
void
set_add_tracks_menu(int enabled)
{
	enable_menu_item(ABN_AddTracksItem, enabled);
}


/*
 * Sets the state of a menu item in the command menu.
 */
void
set_menu_item(int item, int is_set)
{
	ApModifyItemState(&CommandMenu,
			is_set ? AB_ITEM_SET : AB_ITEM_NORMAL,
			item, NULL);
}


/*
 * Enables or disables the playback controls depending on how many songs are
 * in the playlist.
 */
void
set_play_controls()
{
	if (nsongs == 0)
	{
		disable_control(ABW_NextButton, NEXT_BUTTON);
		disable_control(ABW_PrevButton, PREV_BUTTON);
		disable_control(ABW_PlayButton, PLAY_BUTTON);
		disable_control(ABW_StopButton, STOP_BUTTON);
		disable_control(ABW_PauseButton, PAUSE_BUTTON);
		disable_widget(ABW_DelButton);
		enable_menu_item(ABN_SaveItem, 0);
		enable_menu_item(ABN_ClearItem, 0);
	}
	else
	{
		enable_control(ABW_NextButton, NEXT_BUTTON);
		enable_control(ABW_PrevButton, PREV_BUTTON);
		enable_control(ABW_PlayButton, PLAY_BUTTON);
		enable_widget(ABW_DelButton);
		enable_menu_item(ABN_SaveItem, 1);
		enable_menu_item(ABN_ClearItem, 1);

		if (is_playing())
		{
			enable_control(ABW_StopButton, STOP_BUTTON);
			enable_control(ABW_PauseButton, PAUSE_BUTTON);
		}
		else
		{
			disable_control(ABW_StopButton, STOP_BUTTON);
			disable_control(ABW_PauseButton, PAUSE_BUTTON);
		}
	}
}

static void
get_image(PtWidget_t *widget, int index)
{
	PhImage_t	*image;

	PtGetResource(widget, Pt_ARG_LABEL_IMAGE, &image, 0);
	image = PiDuplicateImage(image, Pi_SHMEM);
	if (image == NULL)
		image = PiDuplicateImage(image, 0);
	if (image == NULL)
	{
		perror("Can't get playback control image");
		PtExit(1);
	}

	image->flags &= ~(Ph_RELEASE_IMAGE | Ph_RELEASE_PALETTE |
			Ph_RELEASE_TRANSPARENCY_MASK | Ph_RELEASE_GHOST_BITMAP);
	control_images[index] = image;
}

void
init_play_controls()
{
	get_image(ABW_PrevDisabledImage, PREV_BUTTON);
	get_image(ABW_PrevEnabledImage, PREV_BUTTON + 1);
	get_image(ABW_PlayDisabledImage, PLAY_BUTTON);
	get_image(ABW_PlayEnabledImage, PLAY_BUTTON + 1);
	get_image(ABW_NextDisabledImage, NEXT_BUTTON);
	get_image(ABW_NextEnabledImage, NEXT_BUTTON + 1);
	get_image(ABW_PauseDisabledImage, PAUSE_BUTTON);
	get_image(ABW_PauseEnabledImage, PAUSE_BUTTON + 1);
	get_image(ABW_StopDisabledImage, STOP_BUTTON);
	get_image(ABW_StopEnabledImage, STOP_BUTTON + 1);

	set_play_controls();
}


/*
 * Initializes the loading dialog box.
 */
void
init_loading_dialog(void)
{
	PhPoint_t	*pos;

	ApCreateModule(ABM_LoadingDialog, ABW_PhJuke, NULL);
	PtGetResource(ABW_LoadingDialog, Pt_ARG_POS, &pos, 0);
	loading_pos = *pos;
	loading_is_shown = 1;

	hide_loading_dialog();
}


/*
 * Shows the loading dialog box, updating it with a message and a
 * record count.
 */
void
show_loading_dialog(char *msg, int count)
{
	char		countmsg[100];

	if (! loading_is_shown)
	{
		PtSetResource(ABW_LoadingDialog, Pt_ARG_POS, &loading_pos, 0);
		loading_is_shown = 1;
	}

	sprintf(countmsg, "%d entries...", count);
	set_text(ABW_LoadingLabel, msg);
	set_text(ABW_LoadingEntriesLabel, countmsg);
	PtDamageWidget(ABW_LoadingDialog);

	PtBkgdHandlerProcess();
	if (is_playing())
		usleep(1000);
	PtBkgdHandlerProcess();
}

/*
 * Hides the loading dialog box.
 */
void
hide_loading_dialog(void)
{
	PhPoint_t	pos;

	if (loading_is_shown)
	{
		pos.x = -10000;
		pos.y = -10000;
		PtSetResource(ABW_LoadingDialog, Pt_ARG_POS, &pos, 0);
		PtBkgdHandlerProcess();
		loading_is_shown = 0;
	}
}
