/*
 * Playlist management code.  We keep a list of all songs' data in memory,
 * as well as a list of indexes into the array of songs, which may be in
 * random order if shuffle mode is set.
 */
#include <stdio.h>
#include <errno.h>
#include "ablibs.h"
#include "abimport.h"
#include "abdefine.h"
#include "playlist.h"
#include "gfuncs.h"

#define LIST_BATCH_SIZE 500

SONG	*songs = NULL;
int	*song_indexes;
int	shuffle = 1;
int	repeat = 1;
int	autoclear = 1;
int	nsongs = 0;
int	current_song = -1;		// Position in song_indexes array
int	selected_song = -1;
static int songs_size = 0;
static void *playlist_arena = NULL;


void
clear_list()
{
	int	i;

	arena_destroy(&playlist_arena);

	nsongs = 0;
	current_song = -1;

	PtListDeleteAllItems(ABW_PlayList);
	selected_song = -1;
	set_play_controls();
}


/*
 * Ensures that there's enough space left to add an item to the song list.
 */
static void
ensure_space(void)
{
	int	new_bytes;

	if (nsongs >= songs_size)
	{
		songs_size += 100;

		songs = grealloc(songs, sizeof(SONG) * songs_size);
		song_indexes = grealloc(song_indexes, sizeof(int) * songs_size);
	}

	if (playlist_arena == NULL)
		playlist_arena = arena_create(0);
}

/*
 * Moves the playlist widget to a particular song.
 */
static void
select_in_widget(int songnum)
{
	PtContainerHold(ABW_PlayList);

	if (songnum >= 0)
	{
		// PtListGotoPos will scroll minimally to display the selected
		// item.  We want two items above and below to show if possible
		// so the user can see what's coming up.
		PtListGotoPos(ABW_PlayList, songnum > 1 ? songnum - 1 : 1);
		if (nsongs > 3)
		{
			if (songnum < nsongs - 2)
				PtListGotoPos(ABW_PlayList, songnum + 3);
			else
				PtListGotoPos(ABW_PlayList, nsongs);
		}
		PtListGotoPos(ABW_PlayList, songnum + 1);
		PtListSelectPos(ABW_PlayList, songnum + 1);
	}
	else
	{
		PtListSelectPos(ABW_PlayList, 0);
	}

	PtContainerRelease(ABW_PlayList);

	selected_song = songnum;
}

void
save_list(char *directory, char *filename)
{
	void	*m3u;
	int	i;

	if (nsongs > 250)
		show_loading_dialog("Saving", 0);

	m3u = m3u_create(directory, filename);
	if (m3u == NULL)
	{
		char msg[1000];

		sprintf(msg, "Couldn't open playlist file\n%s",
			sys_errlist[errno]);
		PtMessageBox(NULL,
				"PhJuke Alert",
				msg,
				NULL,
				"OK");
		return;
	}

	for (i = 0; i < nsongs; i++)
	{
		if ((i % 250) == 0 && i > 0)
			show_loading_dialog("Saving", 0);

		if (m3u_write(m3u, songs[i].title, songs[i].filename))
		{
			char msg[1000];

			sprintf(msg, "Couldn't finish writing playlist\n%s",
				sys_errlist[errno]);
			PtMessageBox(NULL,
					"PhJuke Alert",
					msg,
					NULL,
					"OK");
			break;
		}
	}

	m3u_close(m3u);

	hide_loading_dialog();
}


void
add_to_list(char *directory, char *filename, char *title, int update_widget)
{
	char		path[1000];
	const char	*display = title;

	if (title == NULL)
	{
		display = strrchr(filename, '/');
		if (display == NULL)
			display = filename;
		else
			display++;
	}

	ensure_space();
	song_indexes[nsongs] = nsongs;
	songs[nsongs].title = arena_strdup(playlist_arena, (char *)display);

	if (filename[0] != '/')
	{
		sprintf(path, "%s/%s", directory, filename);
		songs[nsongs].filename = arena_strdup(playlist_arena, path);
	}
	else
		songs[nsongs].filename = arena_strdup(playlist_arena, filename);

	if (update_widget)
	{
		PtListAddItems(ABW_PlayList,
			(const char **)&songs[nsongs].title, 1, 0);
	}

	nsongs++;
}


/*
 * Adds an MP3 to the list when only the filename is known.
 *
 * update_widget = 1 to add the file to the end of the widget list.
 */
void
add_filename_to_list(char *filename, int update_widget)
{
	char		path[1000];
	char		title[1000];
	char		*c;
	char		*directory;

	gstrncpy(path, filename, sizeof(path));
	c = strrchr(path, '/');
	if (c == NULL)
	{
		filename = path;
		directory = ".";
	}
	else
	{
		filename = c + 1;
		*c = '\0';
		directory = path;
	}

	strcpy(title, filename);
	c = strrchr(title, '.');
	if (c != NULL)
		*c = '\0';

	fs_to_utf8(title, sizeof(title));
	add_to_list(directory, filename, title, update_widget);
}


/*
 * Figure out the correct track order and the current track depending
 * on whether shuffle is set or not.
 *
 * Updates song_indexes and current_song.
 */
void
order_tracks()
{
	int		current_index;
	long		flags;
	int		i, indx;
	char		*path = NULL;
	const char	**items;

	// If a song is playing, remember which one it is.
	if (current_song >= 0)
		current_index = song_indexes[current_song];
	else
		current_index = -1;

	for (i = 0; i < nsongs; i++)
		song_indexes[i] = i;

	if (shuffle)
	{
		/*
		 * Shuffle mode.  Construct a random-order list, starting
		 * with the currently playing track if any.
		 */
		if (current_index >= 0)
		{
			song_indexes[0] = current_index;
			song_indexes[current_index] = 0;
			current_song = 0;
			i = 1;
		}
		else
			i = 0;

		// Thoroughly mix up the track order.
		for (; i < nsongs; i++)
		{
			register int	temp, pos;

			if (current_index >= 0)
				pos = (random() % (nsongs - 1)) + 1;
			else
				pos = random() % nsongs;
			temp = song_indexes[pos];
			song_indexes[pos] = song_indexes[i];
			song_indexes[i] = temp;

			if ((i % 100) == 0 && i > 0)
				show_loading_dialog("Shuffling", i);
		}
	}
	else
	{
		/*
		 * Normal play mode.  song_indexes[n] = n.
		 */
		current_song = current_index;
	}

	/*
	 * Reconstruct the list widget with the new track order.
	 */
	PtContainerHold(ABW_PlayList);
	PtListDeleteAllItems(ABW_PlayList);
	items = gmalloc(LIST_BATCH_SIZE * sizeof(char *));
	indx = 0;
	for (i = 0; i < nsongs; i++)
	{
		items[indx++] = songs[song_indexes[i]].title;
		if (indx == LIST_BATCH_SIZE)
		{
			show_loading_dialog("Listing", i + 1);
			PtListAddItems(ABW_PlayList, items, indx, 0);
			indx = 0;
		}
	}

	if (indx > 0)
		PtListAddItems(ABW_PlayList, items, indx, 0);
	free(items);

	select_in_widget(current_song);
	PtContainerRelease(ABW_PlayList);
	hide_loading_dialog();
}


/*
 * Moves to the next track in the playlist.
 *
 * Returns:
 *	Filename to play, or NULL if we've reached the end of the list.
 */
char *
next_track()
{
	current_song++;

	/*
	 * If we're at the end and repeat mode is set, rescramble the
	 * playlist.
	 */
	if (current_song == nsongs)
	{
		if (! repeat)
		{
			current_song = -1;
			select_in_widget(current_song);
			return NULL;
		}
		current_song = -1;
		order_tracks();
		current_song = 0;
	}

	select_in_widget(current_song);
	return songs[song_indexes[current_song]].filename;
}


/*
 * Moves to the previous track in the playlist.
 *
 * Returns:
 *	Filename to play, or NULL if we've reached the end of the list.
 */
char *
previous_track()
{
	current_song--;

	/*
	 * If we're at the beginning and repeat mode is set, rescramble the
	 * playlist.
	 */
	if (current_song == -1)
	{
		if (! repeat)
			return NULL;
		current_song = -1;
		order_tracks();
		current_song = nsongs - 1;
	}

	select_in_widget(current_song);
	return songs[song_indexes[current_song]].filename;
}


/*
 * Returns the filename of the current song, or NULL if no song is selected.
 */
char *
current_filename()
{
	if (current_song >= 0)
		return songs[song_indexes[current_song]].filename;
	else
		return NULL;
}


/*
 * Deletes the currently selected song from the playlist.
 */
void
delete_track()
{
	int	delete_pos = selected_song;
	int	delete_index = song_indexes[delete_pos];
	int	i;

	PtListDeleteItemPos(ABW_PlayList, 1, delete_pos + 1);
	nsongs--;
	set_play_controls();

	/*
	 * Remove the song from the internal song list and the index mapping.
	 */
	if (delete_index < nsongs)
	{
		memmove(songs + delete_index,
			songs + delete_index + 1,
			sizeof(SONG) * (nsongs - delete_index));
	}
	if (delete_pos < nsongs)
	{
		memmove(song_indexes + delete_pos,
			song_indexes + delete_pos + 1,
			sizeof(int) * (nsongs - delete_pos));
	}

	for (i = 0; i < nsongs; i++)
		if (song_indexes[i] > delete_index)
			song_indexes[i]--;

	/*
	 * Move the selection forward if possible, backward if not.
	 */
	if (delete_pos < nsongs)
	{
		select_in_widget(delete_pos);
	}
	else if (delete_pos > 0)
	{
		select_in_widget(delete_pos - 1);
	}
	else
		selected_song = -1;

	if (delete_pos == current_song)
	{
		current_song--;
		if (is_playing())
			start_playing(next_track());
		if (current_song == -1 && selected_song == -1)
			selected_song = nsongs - 1;
	}
	else if (delete_pos < current_song)
		current_song--;
}
