/*
 * Code for reading and writing the preferences file.
 */
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "playlist.h"
#include "gfuncs.h"

extern int shuffle, repeat, autoclear, save_on_exit;
extern int use_replaygain;
extern double add_decibels;
extern char *starting_dir;
extern char *native_charset;
extern char *fs_charset;

#define DEFAULT_DEFAULT_LIST "/phjuke.m3u"

char *prefs_file = NULL;	// $HOME/.phjuke if null
char *default_list = NULL;
int default_list_mode = 0;

/*
 * Figures out the preferences filename, if one wasn't specified on the
 * command line.
 */
void
prefs_init()
{
	if (prefs_file == NULL)
	{
		if (getenv("HOME") == NULL)
		{
			fprintf(stderr, "Warning: HOME not set, no preferences "
				"may be loaded or saved\n");
			return;
		}
		else
		{
			char	buf[1000];

			sprintf(buf, "%s/.phjuke", getenv("HOME"));
			prefs_file = gstrdup(buf);
		}
	}
}


/*
 * Reads the preferences file.
 */
void
prefs_read()
{
	char	buf[1000];
	char	*keyword;
	char	*value;
	FILE	*fp;

	if (prefs_file == NULL)
		return;

	fp = fopen(prefs_file, "r");
	if (fp == NULL)
	{
		if (errno != ENOENT)
			perror(prefs_file);
		return;
	}

	while (fgets(buf, sizeof(buf), fp) != NULL)
	{
		strtok(buf, "\r\n");
		if (buf[0] == '\0' || buf[0] == '#')
			continue;

		keyword = strtok(buf, "\t ");
		value = strtok(NULL, "");

		if (! strcasecmp(keyword, "shuffle"))
			shuffle = (value == NULL) ? 1 : atoi(value);
		else if (! strcasecmp(keyword, "repeat"))
			repeat = (value == NULL) ? 1 : atoi(value);
		else if (! strcasecmp(keyword, "autoclear"))
			autoclear = (value == NULL) ? 1 : atoi(value);
		else if (! strcasecmp(keyword, "covers") && value != NULL &&
			atoi(value) != 0)
		{
			text_to_cover_filenames("cover.jpg,cover.gif,cover.bmp");
		}
		else if (! strcasecmp(keyword, "cover-file"))
			add_cover_filename(value);
		else if (! strcasecmp(keyword, "directory"))
		{
			if (starting_dir == NULL && value != NULL)
				starting_dir = gstrdup(value);
		}
		else if (! strcasecmp(keyword, "fs-charset") && value != NULL)
			fs_charset = gstrdup(value);
		else if (! strcasecmp(keyword, "data-charset") && value != NULL)
			native_charset = gstrdup(value);
		else if (! strcasecmp(keyword, "port") && value != NULL)
			network_init(atoi(value));
		else if (! strcasecmp(keyword, "map-dir") && value != NULL)
			add_map_dir(value);
		else if (! strcasecmp(keyword, "default-list") &&
		         value != NULL && default_list == NULL)
		{
			default_list = gstrdup(value);
		}
		else if (! strcasecmp(keyword, "save-on-exit"))
			save_on_exit = (value == NULL) ? 1 : atoi(value);
		else if (! strcasecmp(keyword, "default-list-mode"))
			default_list_mode = (value == NULL) ? 0 : atoi(value);
		else if (! strcasecmp(keyword, "replay-gain"))
			use_replaygain = (value == NULL) ? 0 : atoi(value);
		else if (! strcasecmp(keyword, "add-decibels"))
			add_decibels = (value == NULL) ? 0 : atof(value);
	}

	fclose(fp);

	if (default_list == NULL)
	{
		char	*home = getenv("HOME");

		if (home == NULL || ! strcmp(home, "/"))
			default_list = gstrdup(DEFAULT_DEFAULT_LIST);
		else
		{
			default_list = gmalloc(strlen(home) +
						sizeof(DEFAULT_DEFAULT_LIST));
			strcpy(default_list, home);
			strcat(default_list, DEFAULT_DEFAULT_LIST);
		}
	}
}


/*
 * Writes the preferences file.
 */
void
prefs_write()
{
	char	buf[1000];
	char	*keyword;
	char	*value;
	char	*cover_files = cover_filenames_to_text();
	FILE	*fp;

	if (prefs_file == NULL)
		return;

	fp = fopen(prefs_file, "w");
	if (fp == NULL)
	{
		perror(prefs_file);
		return;
	}

	fprintf(fp, "shuffle %d\n", shuffle);
	fprintf(fp, "repeat %d\n", repeat);
	fprintf(fp, "autoclear %d\n", autoclear);
	if (cover_files != NULL && cover_files[0] != '\0')
		fprintf(fp, "cover-file %s\n", cover_files);
	fprintf(fp, "save-on-exit %d\n", save_on_exit);
	fprintf(fp, "default-list-mode %d\n", default_list_mode);
	fprintf(fp, "directory %s\n", starting_dir);
	fprintf(fp, "default-list %s\n", default_list);
	fprintf(fp, "add-decibels %f\n", add_decibels);
	fprintf(fp, "replay-gain %d\n", use_replaygain);

	fprintf(fp, "fs-charset %s\n", fs_charset);
	fprintf(fp, "data-charset %s\n", native_charset);

	print_map_dirs(fp);

	if (get_port() > 0)
		fprintf(fp, "port %d\n", get_port());
	fclose(fp);

	gfree(cover_files);
}
