/*
 * Manipulation of pipe-delimited textfile playlists, as exported by
 * Media Jukebox (using the "export as ASCII file" command).
 *
 * The first line of the file is the field specifiction.  We care about
 * the "Name", "Artist", and "Location" fields.
 */
#include <stdio.h>
#include <string.h>
#include "playlist.h"
#include "gfuncs.h"
#include "decoder.h"

#define FIELD_ARTIST	0
#define FIELD_ALBUM	1
#define FIELD_NAME	2
#define FIELD_LOCATION	3
#define NFIELDS		4

static char *field_names[NFIELDS] = { "Artist", "Album", "Name", "Location" };

/*
 * Extracts the next field from a pipe-delimited, newline-terminated line.
 */
static char *next_field(char *str)
{
	char	*field;

	// Skip past the current field.
	while (*str != '|' && *str != '\0')
		str++;
	// Jump over the delimiter.
	str++;

	field = str;

	// Find the end delimiter.
	while (*str != '|' && *str != '\0' && *str != '\r' && *str != '\n')
		str++;

	// The line ends with an empty field.
	if (*str == '\r' || *str == '\n' || *str == '\0')
		return NULL;

	*str = '\0';
	return field;
}

/*
 * Reads a pipe-separated-value playlist file and adds all its records to the
 * playlist.
 *
 * Returns:
 *	0	Playlist loaded successfully
 *	1	Couldn't load list.
 */
int
load_txt(char *filename, int update_widget)
{
	char	*c;
	char	directory[1000];
	char	title[1000];
	char	line[1000];
	char	*field;
	int	field_nums[NFIELDS];
	char	*field_values[NFIELDS];
	FILE	*fp;
	int	count = 0;
	int	got_fields;
	int	field_num;
	int	i;

	title[0] = '\0';

	c = strrchr(filename, '/');
	if (c == NULL)
		strcpy(directory, ".");
	else
	{
		*c = '\0';
		strcpy(directory, filename);
		*c = '/';
	}

	fp = fopen(filename, "r");
	if (fp == NULL)
	{
		perror(filename);
		return 1;
	}

	/*
	 * Figure out the field order based on the header line.  Since
	 * the order can change, look for the field names we care about
	 * and note which field number each one is.
	 */
	if (fgets(line, sizeof(line), fp) != NULL)
	{
		for (field = line, field_num = 0;
		     field != NULL;
		     field = next_field(field), field_num++)
		{
			for (i = 0; i < NFIELDS; i++)
				if (! strcmp(field, field_names[i]))
					field_nums[i] = field_num;
		}
	}

	while (fgets(line, sizeof(line), fp) != NULL)
	{
		if (line[0] != '|')
			continue;

		/*
		 * Extract the relevant fields from the line.
		 */
		for (i = 0; i < NFIELDS; i++)
			field_values[i] = "";
		for (field = line, field_num = 0, got_fields = 0;
		     field != NULL && got_fields < NFIELDS;
		     field = next_field(field), field_num++)
		{
			for (i = 0; i < NFIELDS; i++)
				if (field_num == field_nums[i])
				{
					field_values[i] = field;
					got_fields++;
					break;
				}
		}

		if (! find_decoder(field_values[FIELD_LOCATION]))
			continue;

		if ((++count % 250) == 0)
			show_loading_dialog("Loading", count);

		if (field_values[FIELD_NAME][0] == '\0') // no track title
		{
			// Construct title out of filename
			c = strrchr(field_values[FIELD_LOCATION], '/');
			if (c == NULL)
				strcpy(title, line);
			else
				strcpy(title, c + 1);

			// Strip off extension
			c = strrchr(title, '.');
			if (c != NULL)
				*c = '\0';
		}
		else
		{
			sprintf(title, "%s - %s",
				field_values[FIELD_ARTIST],
				field_values[FIELD_NAME]);
		}

		native_to_utf8(title, sizeof(title));

		/* Convert Windows relative paths to UNIX format */
		for (c = field_values[FIELD_LOCATION]; *c != '\0'; c++)
			if (*c == '\\')
				*c = '/';

		/* Convert to local mount points */
		map_dir(field_values[FIELD_LOCATION]);

		add_to_list(directory, field_values[FIELD_LOCATION], title,
				update_widget);
	}

	fclose(fp);
	set_play_controls();

	return (count == 0);
}


/*
 * Checks a .txt file to see if it's really a Media Jukebox playlist.
 *
 * Input:
 *	Filename to check
 * Returns:
 *	Nonzero if this is a MJ playlist.
 */
int
check_txt(const char *filename)
{
	FILE	*fp;
	char	line[500];
	int	is_mj = 0;

	if (! endswith(filename, ".txt"))
		return 0;
	
	fp = fopen(filename, "r");
	if (fp == NULL)
		perror(filename);
	else
	{
		if (fgets(line, sizeof(line), fp) != NULL)
		{
			if (line[0] == '|' && strstr(line, "|Name|") &&
			    strstr(line, "|Artist|") &&
			    strstr(line, "|Location|"))
			{
				is_mj = 1;
			}
		}

		fclose(fp);
	}

	return is_mj;
}
